import datetime
import unittest

from walrus.streams import TimeSeries
from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


class TestTimeSeries(WalrusTestCase):
    def setUp(self):
        super(TestTimeSeries, self).setUp()
        for key in ('sa', 'sb', 'sc'):
            db.delete('key')

        self.ts = TimeSeries(db, 'cgabc', {'sa': '0', 'sb': '0', 'sc': '0'},
                             consumer='ts1')

    def _create_test_data(self):
        start = datetime.datetime(2018, 1, 1)
        id_list = []
        keys = ('sa', 'sb', 'sc')
        for i in range(0, 10):
            tskey = getattr(self.ts, keys[i % 3])
            ts = start + datetime.timedelta(days=i)
            id_list.append(tskey.add({'k': '%s-%s' % (keys[i % 3], i)}, id=ts))
        return id_list

    def assertMessages(self, results, expected_ids):
        rdata = [(r.stream, r.timestamp, r.data) for r in results]
        streams = ('sa', 'sb', 'sc')
        edata = [(streams[i % 3], datetime.datetime(2018, 1, i + 1),
                 {'k': '%s-%s' % (streams[i % 3], i)}) for i in expected_ids]
        self.assertEqual(rdata, edata)

    def test_timeseries_ranges(self):
        docids = self._create_test_data()

        self.ts.create()
        self.assertMessages(self.ts.sa.range(), [0, 3, 6, 9])
        self.assertMessages(self.ts.sb.range(), [1, 4, 7])
        self.assertMessages(self.ts.sc.range(), [2, 5, 8])
        self.assertMessages(self.ts.sc.range(count=2), [2, 5])

        self.assertMessages(self.ts.sa[:docids[4]], [0, 3])
        self.assertMessages(self.ts.sb[:docids[4]], [1, 4])
        self.assertMessages(self.ts.sa[docids[4]:], [6, 9])
        self.assertMessages(self.ts.sb[docids[4]:], [4, 7])

        self.assertMessages([self.ts.sa.get(docids[6])], [6])
        self.assertMessages([self.ts.sa.get(docids[9])], [9])
        self.assertMessages([self.ts.sc.get(docids[5])], [5])
        self.assertTrue(self.ts.sa.get(docids[5]) is None)
        self.assertTrue(self.ts.sb.get(docids[5]) is None)

        # Trim sa down to 2 items.
        self.assertEqual(self.ts.sa.trim(2, False), 2)
        self.assertMessages(self.ts.sa.range(), [6, 9])
        self.assertMessages(self.ts.sa.range(count=1), [6])
        self.assertMessages(self.ts.streams['sa'].range(count=1), [6])

    def test_timeseries_read(self):
        self._create_test_data()
        self.ts.create()
        self.assertMessages(self.ts.read(count=1), [0, 1, 2])
        self.assertMessages(self.ts.read(count=1), [3, 4, 5])
        self.assertMessages(self.ts.read(count=2), [6, 7, 8, 9])
        self.assertEqual(self.ts.read(), [])

        # Trim the 0-th item off of sa and reset all streams.
        self.ts.sa.trim(3, False)
        self.ts.reset()
        self.assertMessages(self.ts.read(), list(range(1, 10)))

        # Trim the first two items from sc (2 and 5), then set the date so
        # we've read the first item from each queue. Next read will be 4.
        self.ts.sc.trim(1, False)
        self.ts.sb.delete(datetime.datetime(2018, 1, 8))  # Delete item 7.
        self.ts.set_id(datetime.datetime(2018, 1, 4))
        self.assertMessages(self.ts.read(), [4, 6, 8, 9])

    def test_adding(self):
        self._create_test_data()
        self.ts.create()
        resp = self.ts.set_id()
        self.assertEqual(resp, {'sa': True, 'sb': True, 'sc': True})

        # We can add another record to the max ts if we increment the seq.
        r = self.ts.sa.add({'k': 'sa-10'}, (datetime.datetime(2018, 1, 10), 1))
        self.assertEqual(r, (datetime.datetime(2018, 1, 10), 1))

        r = self.ts.sb.add({'k': 'sb-11'}, (datetime.datetime(2018, 1, 10), 2))
        self.assertEqual(r, (datetime.datetime(2018, 1, 10), 2))

        self.assertEqual(len(self.ts.sa), 5)
        self.assertEqual(len(self.ts.sb), 4)
        self.assertEqual(len(self.ts.sc), 3)

        # Read the newly-added records.
        r10, r11 = self.ts.read()
        self.assertEqual(r10.timestamp, datetime.datetime(2018, 1, 10))
        self.assertEqual(r10.sequence, 1)
        self.assertEqual(r10.stream, 'sa')
        self.assertEqual(r10.data, {'k': 'sa-10'})
        self.assertEqual(r11.timestamp, datetime.datetime(2018, 1, 10))
        self.assertEqual(r11.sequence, 2)
        self.assertEqual(r11.stream, 'sb')
        self.assertEqual(r11.data, {'k': 'sb-11'})

    def test_timeseries_stream_read(self):
        self._create_test_data()
        self.ts.create()

        # Read two from sa, one from sc, then read 2x from all. Messages that
        # were read will not be re-read.
        self.assertMessages(self.ts.sa.read(count=2), [0, 3])
        self.assertMessages(self.ts.sc.read(count=1), [2])
        self.assertMessages(self.ts.read(count=2), [1, 4, 5, 6, 8, 9])
        self.assertMessages(self.ts.read(count=1), [7])
        for s in (self.ts.sa, self.ts.sb, self.ts.sc):
            self.assertEqual(s.read(), [])

        # Re-set the ID of stream b. Other streams are unaffected, so we just
        # re-read items from stream b.
        self.ts.sb.set_id(datetime.datetime(2018, 1, 4))
        self.assertMessages(self.ts.read(), [4, 7])

        # Re-set the ID of stream a and trim.
        self.ts.sa.set_id('0')
        self.ts.sa.trim(2, False)
        self.assertMessages(self.ts.read(), [6, 9])

    def test_ack_claim_pending(self):
        self._create_test_data()
        self.ts.create()
        ts1 = self.ts
        ts2 = ts1.consumer('ts2')

        # Read items 0, 1, 3, and 4.
        self.assertMessages(ts1.sa.read(1), [0])
        self.assertMessages(ts2.sb.read(2), [1, 4])
        self.assertMessages(ts2.sa.read(1), [3])

        def assertPending(resp, expected):
            clean = [(r[0][0], r[1], r[3]) for r in resp]
            self.assertEqual(clean, expected)

        # Check pending status. sa was read by ts1 first, then ts2.
        assertPending(ts1.sa.pending(), [
            (datetime.datetime(2018, 1, 1), 'ts1', 1),
            (datetime.datetime(2018, 1, 4), 'ts2', 1)])
        assertPending(ts1.sa.pending(consumer='ts1'), [
            (datetime.datetime(2018, 1, 1), 'ts1', 1)])
        assertPending(ts1.sa.pending(consumer='ts2'), [
            (datetime.datetime(2018, 1, 4), 'ts2', 1)])

        # sb was read by ts2 only.
        assertPending(ts1.sb.pending(), [
            (datetime.datetime(2018, 1, 2), 'ts2', 1),
            (datetime.datetime(2018, 1, 5), 'ts2', 1)])

        # Acknowledge receipt. Although we read the Jan 4th item from "sa"
        # using ts2, we can still acknowledge it from ts1.
        self.assertEqual(ts1.sa.ack(datetime.datetime(2018, 1, 4)), 1)
        self.assertEqual(ts2.sb.ack(datetime.datetime(2018, 1, 2),
                                    datetime.datetime(2018, 1, 5)), 2)

        # Verify pending removed.
        assertPending(ts1.sa.pending(), [
            (datetime.datetime(2018, 1, 1), 'ts1', 1)])
        assertPending(ts2.sb.pending(), [])

        # Claim the first message for consumer ts2.
        resp = ts2.sa.claim(datetime.datetime(2018, 1, 1))
        self.assertMessages(resp, [0])

        # Pending is now marked for ts2, ack'd, and removed.
        assertPending(ts1.sa.pending(), [
            (datetime.datetime(2018, 1, 1), 'ts2', 2)])
        self.assertEqual(ts2.sa.ack(datetime.datetime(2018, 1, 1)), 1)
        assertPending(ts2.sa.pending(), [])

        # Read the rest from consumer ts2 and verify pending.
        self.assertMessages(ts2.read(), [2, 5, 6, 7, 8, 9])
        assertPending(ts2.sa.pending(), [
            (datetime.datetime(2018, 1, 7), 'ts2', 1),
            (datetime.datetime(2018, 1, 10), 'ts2', 1)])
        assertPending(ts2.sb.pending(), [
            (datetime.datetime(2018, 1, 8), 'ts2', 1)])
        assertPending(ts2.sc.pending(), [
            (datetime.datetime(2018, 1, 3), 'ts2', 1),
            (datetime.datetime(2018, 1, 6), 'ts2', 1),
            (datetime.datetime(2018, 1, 9), 'ts2', 1)])

        # Claim the records in sc for ts1.
        resp = ts1.sc.claim(
            datetime.datetime(2018, 1, 3),
            datetime.datetime(2018, 1, 6),
            datetime.datetime(2018, 1, 9))
        self.assertMessages(resp, [2, 5, 8])



import os
import unittest

from walrus import Database


HOST = os.environ.get('WALRUS_REDIS_HOST') or '127.0.0.1'
PORT = os.environ.get('WALRUS_REDIS_PORT') or 6379

db = Database(host=HOST, port=PORT, db=15)


REDIS_VERSION = None


class WalrusTestCase(unittest.TestCase):
    def setUp(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def tearDown(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def assertList(self, values, expected):
        values = list(values)
        self.assertEqual(len(values), len(expected))
        for value, item in zip(values, expected):
            self.assertEqual(value, item)



import time

from walrus.rate_limit import RateLimitException
from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


class TestRateLimit(WalrusTestCase):
    def setUp(self):
        super(TestRateLimit, self).setUp()
        # Limit to 5 events per second.
        self.rl = self.get_rate_limit('test-rl', 5, 1)

    def get_rate_limit(self, key, limit, per):
        return db.rate_limit(key, limit, per)

    def test_rate_limit(self):
        for i in range(5):
            self.assertFalse(self.rl.limit('k1'))

        for i in range(3):
            self.assertTrue(self.rl.limit('k1'))

        self.assertFalse(self.rl.limit('k2'))

    def test_rate_limit_rollover(self):
        rl = self.get_rate_limit('test-rl2', 3, 100)
        container = db.List('test-rl2:k1')

        now = time.time()
        past = now - 101

        # Simulate two events.
        container.extend([now, now])

        # Third event goes through OK.
        self.assertFalse(rl.limit('k1'))

        # Fourth event is rate-limited.
        self.assertTrue(rl.limit('k1'))

        # There are three timestamps in the container.
        self.assertEqual(len(container), 3)

        # Hand modify the oldest timestamp to appear as if it happened over
        # 100 seconds ago.
        container[-1] = past

        # We can again perform an action.
        self.assertFalse(rl.limit('k1'))

        # We once again have 3 items all within the last 100 seconds, so we
        # are rate-limited.
        self.assertTrue(rl.limit('k1'))

        # There are only 3 items in the container.
        self.assertEqual(len(container), 3)

        # The oldest item is the 2nd we added at the beginning of the test.
        self.assertEqual(float(container[-1]), now)

        # Remove an item and make the 2nd timestamp (oldest) in the past. This
        # gives us 2 actions.
        container.popright()
        container[-1] = past

        self.assertFalse(rl.limit('k1'))
        self.assertFalse(rl.limit('k1'))
        self.assertTrue(rl.limit('k1'))

    def test_decorator(self):
        rl = self.get_rate_limit('test-rl2', 3, 100)
        container = db.List('test-rl2:fake-key')

        def key_fn(*args, **kwargs):
            return 'fake-key'

        @rl.rate_limited(key_function=key_fn)
        def do_test():
            return 'OK'

        now = time.time()
        container.extend([now, now])

        self.assertEqual(do_test(), 'OK')
        self.assertRaises(RateLimitException, do_test)

        container.popright()
        container[-1] = now - 101

        self.assertEqual(do_test(), 'OK')
        self.assertEqual(do_test(), 'OK')
        self.assertRaises(RateLimitException, do_test)


class TestRateLimitLua(TestRateLimit):
    def get_rate_limit(self, key, limit, per):
        return db.rate_limit_lua(key, limit, per)



import os
import re
import sys


PY3 = sys.version_info[0] == 3

if PY3:
    unicode_type = str
    basestring_type = (str, bytes)
    def exception_message(exc):
        return exc.args[0]
else:
    unicode_type = unicode
    basestring_type = basestring
    def exception_message(exc):
        return exc.message


def encode(s):
    return s.encode('utf-8') if isinstance(s, unicode_type) else s


def decode(s):
    return s.decode('utf-8') if isinstance(s, bytes) else s


def decode_dict(d):
    accum = {}
    for key in d:
        accum[decode(key)] = decode(d[key])
    return accum


def safe_decode_list(l):
    return [i.decode('raw_unicode_escape') if isinstance(i, bytes) else i
            for i in l]


def decode_dict_keys(d):
    accum = {}
    for key in d:
        accum[decode(key)] = d[key]
    return accum


def make_python_attr(s):
    if isinstance(s, bytes):
        s = decode(s)
    s = re.sub(r'[^\w]+', '_', s)
    if not s:
        raise ValueError('cannot construct python identifer from "%s"' % s)
    if s[0].isdigit():
        s = '_' + s
    return s.lower()


class memoize(dict):
    def __init__(self, fn):
        self._fn = fn

    def __call__(self, *args):
        return self[args]

    def __missing__(self, key):
        result = self[key] = self._fn(*key)
        return result


@memoize
def load_stopwords(stopwords_file):
    path, filename = os.path.split(stopwords_file)
    if not path:
        path = os.path.dirname(__file__)
    filename = os.path.join(path, filename)
    if not os.path.exists(filename):
        return

    with open(filename) as fh:
        return fh.read()



#coding:utf-8
from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


class TestSearchIndex(WalrusTestCase):
    def test_search_index(self):
        phrases = [
            ('A faith is a necessity to a man. Woe to him who believes in '
             'nothing.'),
            ('All who call on God in true faith, earnestly from the heart, '
             'will certainly be heard, and will receive what they have asked '
             'and desired.'),
            ('Be faithful in small things because it is in them that your '
             'strength lies.'),
            ('Faith consists in believing when it is beyond the power of '
             'reason to believe.'),
            ('Faith has to do with things that are not seen and hope with '
             'things that are not at hand.')]

        index = db.Index('test-index')

        for idx, message in enumerate(phrases):
            index.add('doc-%s' % idx, message)

        def assertDocs(query, indexes):
            results = [doc['content'] for doc in index.search(query)]
            self.assertEqual(results, [phrases[i] for i in indexes])

        assertDocs('faith', [0, 2, 3, 4, 1])
        assertDocs('faith man', [0])
        assertDocs('things', [4, 2])
        assertDocs('blah', [])

    def test_add_remove_update(self):
        data = [
            ('huey cat', {'type': 'cat', 'color': 'white'}),
            ('zaizee cat cat', {'type': 'cat', 'color': 'gray'}),
            ('mickey dog', {'type': 'dog', 'color': 'black'}),
            ('beanie cat', {'type': 'cat', 'color': 'gray'}),
        ]

        idx = db.Index('test-index')
        for i, (content, metadata) in enumerate(data):
            idx.add(str(i), content, **metadata)

        huey, = idx.search('huey')
        self.assertEqual(huey, {
            'content': 'huey cat',
            'type': 'cat',
            'color': 'white'})

        self.assertEqual([d['content'] for d in idx.search('cat')],
                         ['zaizee cat cat', 'huey cat', 'beanie cat'])

        idx.remove('3')  # Poor beanie :(
        zaizee, huey = idx.search('cat')
        self.assertEqual(zaizee['content'], 'zaizee cat cat')
        self.assertEqual(huey['content'], 'huey cat')

        self.assertRaises(KeyError, idx.remove, '3')

        idx.update('1', 'zaizee cat', {'type': 'kitten'})
        idx.replace('0', 'huey baby cat', {'type': 'kitten'})

        zaizee, huey = idx.search('cat')
        self.assertEqual(zaizee['content'], 'zaizee cat')
        self.assertEqual(zaizee['type'], 'kitten')
        self.assertEqual(zaizee['color'], 'gray')

        self.assertEqual(huey['content'], 'huey baby cat')
        self.assertEqual(huey['type'], 'kitten')
        self.assertTrue('color' not in huey)

        zaizee, huey = idx.search_items('cat')
        self.assertEqual(zaizee[0], '1')
        self.assertEqual(zaizee[1]['content'], 'zaizee cat')
        self.assertEqual(huey[0], '0')
        self.assertEqual(huey[1]['content'], 'huey baby cat')

    def test_search_phonetic(self):
        data = (
            ('pf', 'python and flask'),
            ('lcp', 'learning cython programming'),
            ('lwd', 'learning web development with flask'),
            ('pwd', 'python web development'))
        data_dict = dict(data)
        idx = db.Index('test-index', metaphone=True)
        for key, content in data:
            idx.add(key, content)

        def assertResults(query, keys):
            result = idx.search(query)
            self.assertEqual([doc['content'] for doc in result],
                             [data_dict[key] for key in keys])

        assertResults('flasck', ['pf', 'lwd'])
        assertResults('pythonn', ['pf', 'pwd'])
        assertResults('sithon', ['lcp'])
        assertResults('webb development', ['pwd', 'lwd'])

        assertResults('sithon OR (flasck AND pythonn)', ['pf', 'lcp'])
        assertResults('garbage', [])

    def test_search_parser(self):
        messages = [
            'foo green',
            'bar green',
            'baz blue',
            'nug blue',
            'nize yellow',
            'huey greener',
            'mickey greens',
            'zaizee',
        ]
        index = db.Index('testing')

        for idx, message in enumerate(messages):
            index.add(str(idx), message)

        def assertMatches(query, expected):
            results = [doc['content'] for doc in index.search(query)]
            self.assertEqual(results, expected)

        assertMatches('foo', ['foo green'])
        assertMatches('foo OR baz', ['foo green', 'baz blue'])
        assertMatches('green OR blue', [
            'foo green',
            'bar green',
            'baz blue',
            'nug blue',
            'mickey greens',
        ])

        assertMatches('green AND (bar OR mickey OR nize)', [
            'bar green',
            'mickey greens',
        ])
        assertMatches('zaizee OR (blue AND nug) OR (green AND bar)', [
            'bar green',
            'nug blue',
            'zaizee',
        ])
        assertMatches('(blue AND (baz OR (nug OR huey OR mickey))', [
            'baz blue',
            'nug blue',
        ])
        assertMatches(
            '(blue OR foo) AND (green OR (huey OR (baz AND mickey)))',
            ['foo green'])

        assertMatches('(green AND nug) OR (blue AND bar)', [])
        assertMatches('nuglet', [])
        assertMatches('foobar', [])
        assertMatches('foo"bar green', ['foo green'])

        results = [doc['content'] for doc in index.search('')]
        self.assertEqual(sorted(results), sorted(messages))

    def test_unicode_handling(self):
        index = db.Index('testing', stemmer=False)
        index.add('1', u'сколько лет этому безумному моржу', {'val': 'age'})
        index.add('2', u'во сколько морж ложится спать', val='sleep')
        index.add('3', u'Вы знаете какие-нибудь хорошие истории с моржами',
                  val='stories')
        self.assertEqual([r['val'] for r in index.search(u'морж')], ['sleep'])



from walrus.containers import *
from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


class TestWalrus(WalrusTestCase):
    def test_get_key(self):
        h = db.Hash('h1')
        h['hk1'] = 'v1'

        l = db.List('l1')
        l.append('i1')

        s = db.Set('s1')
        s.add('k1')

        zs = db.ZSet('z1')
        zs.add({'i1': 1., 'i2': 2.})

        h_db = db.get_key('h1')
        self.assertTrue(isinstance(h_db, Hash))
        self.assertEqual(h_db['hk1'], b'v1')

        l_db = db.get_key('l1')
        self.assertTrue(isinstance(l_db, List))
        self.assertEqual(l_db[0], b'i1')

        s_db = db.get_key('s1')
        self.assertTrue(isinstance(s_db, Set))
        self.assertEqual(s_db.members(), set((b'k1',)))

        z_db = db.get_key('z1')
        self.assertTrue(isinstance(z_db, ZSet))
        self.assertEqual(z_db.score('i1'), 1.)

    def test_atomic(self):
        def assertDepth(depth):
            self.assertEqual(len(db._transaction_local.pipes), depth)

        assertDepth(0)
        with db.atomic() as p1:
            assertDepth(1)
            with db.atomic() as p2:
                assertDepth(2)
                with db.atomic() as p3:
                    assertDepth(3)
                    p3.pipe.set('k3', 'v3')

                assertDepth(2)
                self.assertEqual(db['k3'], b'v3')

                p2.pipe.set('k2', 'v2')

            assertDepth(1)
            self.assertEqual(db['k3'], b'v3')
            self.assertEqual(db['k2'], b'v2')
            p1.pipe.set('k1', 'v1')

        assertDepth(0)
        self.assertEqual(db['k1'], b'v1')
        self.assertEqual(db['k2'], b'v2')
        self.assertEqual(db['k3'], b'v3')

    def test_atomic_exception(self):
        def do_atomic(k, v, exc=False):
            with db.atomic() as a:
                a.pipe.set(k, v)
                if exc:
                    raise TypeError('foo')

        do_atomic('k', 'v')
        self.assertEqual(db['k'], b'v')

        self.assertRaises(TypeError, do_atomic, 'k2', 'v2', True)
        self.assertRaises(KeyError, lambda: db['k2'])
        self.assertEqual(db._transaction_local.pipe, None)

        # Try nested failure.
        with db.atomic() as outer:
            outer.pipe.set('k2', 'v2')
            self.assertRaises(TypeError, do_atomic, 'k3', 'v3', True)

            # Only this will be set.
            outer.pipe.set('k4', 'v4')

        self.assertTrue(db._transaction_local.pipe is None)
        self.assertEqual(db['k2'], b'v2')
        self.assertRaises(KeyError, lambda: db['k3'])
        self.assertEqual(db['k4'], b'v4')

    def test_clear_transaction(self):
        with db.atomic() as a1:
            a1.pipe.set('k1', 'v1')
            with db.atomic() as a2:
                a2.pipe.set('k2', 'v2')
                a2.clear()

        self.assertEqual(db['k1'], b'v1')
        self.assertRaises(KeyError, lambda: db['k2'])

        with db.atomic() as a1:
            a1.pipe.set('k3', 'v3')
            with db.atomic() as a2:
                self.assertRaises(KeyError, lambda: db['k3'])

                a2.pipe.set('k4', 'v4')
                a2.clear()

            a1.pipe.set('k5', 'v5')

        self.assertEqual(db['k3'], b'v3')
        self.assertRaises(KeyError, lambda: db['k4'])
        self.assertEqual(db['k5'], b'v5')

        self.assertTrue(db._transaction_local.pipe is None)

    def test_cas(self):
        db['k1'] = 'v1'
        self.assertTrue(db.cas('k1', 'v1', 'v1-x'))
        self.assertFalse(db.cas('k1', 'v1-z', 'v1-y'))

        self.assertEqual(db['k1'], b'v1-x')
        self.assertTrue(db.cas('k1', 'v1-', 'v2'))
        self.assertFalse(db.cas('k1', 'v1', 'v3'))
        self.assertEqual(db['k1'], b'v2')



from setuptools import find_packages
from setuptools import setup

setup(name='walrus',
      install_requires=['redis>=3.0.0'],
      packages=find_packages(),
      package_data={'walrus': ['scripts/*', 'stopwords.txt']})



import hashlib
import operator
import struct
try:
    from functools import reduce
except ImportError:
    pass
from functools import wraps
try:
    from redis.exceptions import ResponseError
except ImportError:
    ResponseError = Exception

from walrus.utils import basestring_type
from walrus.utils import decode as _decode
from walrus.utils import decode_dict
from walrus.utils import encode
from walrus.utils import exception_message
from walrus.utils import make_python_attr
from walrus.utils import safe_decode_list


def chainable_method(fn):
    @wraps(fn)
    def inner(self, *args, **kwargs):
        fn(self, *args, **kwargs)
        return self
    return inner


class Sortable(object):
    def sort(self, pattern=None, limit=None, offset=None, get_pattern=None,
             ordering=None, alpha=True, store=None):
        if limit or offset:
            offset = offset or 0
        return self.database.sort(
            self.key,
            start=offset,
            num=limit,
            by=pattern,
            get=get_pattern,
            desc=ordering in ('DESC', 'desc'),
            alpha=alpha,
            store=store)


class Container(object):
    """
    Base-class for rich Redis object wrappers.
    """
    def __init__(self, database, key):
        self.database = database
        self.key = key

    def expire(self, ttl=None):
        """
        Expire the given key in the given number of seconds.
        If ``ttl`` is ``None``, then any expiry will be cleared
        and key will be persisted.
        """
        if ttl is not None:
            self.database.expire(self.key, ttl)
        else:
            self.database.persist(self.key)

    def pexpire(self, ttl=None):
        """
        Expire the given key in the given number of milliseconds.
        If ``ttl`` is ``None``, then any expiry will be cleared
        and key will be persisted.
        """
        if ttl is not None:
            self.database.pexpire(self.key, ttl)
        else:
            self.database.persist(self.key)

    def dump(self):
        """
        Dump the contents of the given key using Redis' native
        serialization format.
        """
        return self.database.dump(self.key)

    @chainable_method
    def clear(self):
        """
        Clear the contents of the container by deleting the key.
        """
        self.database.delete(self.key)


class Hash(Container):
    """
    Redis Hash object wrapper. Supports a dictionary-like interface
    with some modifications.

    See `Hash commands <http://redis.io/commands#hash>`_ for more info.
    """
    def __repr__(self):
        l = len(self)
        if l > 5:
            # Get a few keys.
            data = self.database.hscan(self.key, count=5)
        else:
            data = self.as_dict()
        return '<Hash "%s": %s>' % (self.key, data)

    def __getitem__(self, item):
        """
        Retrieve the value at the given key. To retrieve multiple
        values at once, you can specify multiple keys as a tuple or
        list:

        .. code-block:: python

            hsh = db.Hash('my-hash')
            first, last = hsh['first_name', 'last_name']
        """
        if isinstance(item, (list, tuple)):
            return self.database.hmget(self.key, item)
        else:
            return self.database.hget(self.key, item)

    def get(self, key, fallback=None):
        val = self.database.hget(self.key, key)
        return val if val is not None else fallback

    def __setitem__(self, key, value):
        """Set the value of the given key."""
        return self.database.hset(self.key, key, value)

    def __delitem__(self, key):
        """Delete the key from the hash."""
        return self.database.hdel(self.key, key)

    def __contains__(self, key):
        """
        Return a boolean valud indicating whether the given key
        exists.
        """
        return self.database.hexists(self.key, key)

    def __len__(self):
        """Return the number of keys in the hash."""
        return self.database.hlen(self.key)

    def _scan(self, *args, **kwargs):
        return self.database.hscan_iter(self.key, *args, **kwargs)

    def __iter__(self):
        """Iterate over the items in the hash."""
        return iter(self._scan())

    def search(self, pattern, count=None):
        """
        Search the keys of the given hash using the specified pattern.

        :param str pattern: Pattern used to match keys.
        :param int count: Limit number of results returned.
        :returns: An iterator yielding matching key/value pairs.
        """
        return self._scan(match=pattern, count=count)

    def keys(self):
        """Return the keys of the hash."""
        return self.database.hkeys(self.key)

    def values(self):
        """Return the values stored in the hash."""
        return self.database.hvals(self.key)

    def items(self, lazy=False):
        """
        Like Python's ``dict.items()`` but supports an optional
        parameter ``lazy`` which will return a generator rather than
        a list.
        """
        if lazy:
            return self._scan()
        else:
            return list(self)

    def setnx(self, key, value):
        """
        Set ``key`` to ``value`` if ``key`` does not exist.

        :returns: True if successfully set or False if the key already existed.
        """
        return bool(self.database.hsetnx(self.key, key, value))

    @chainable_method
    def update(self, __data=None, **kwargs):
        """
        Update the hash using the given dictionary or key/value pairs.
        """
        if __data is None:
            __data = kwargs
        else:
            __data.update(kwargs)
        return self.database.hset(self.key, mapping=__data)

    def incr(self, key, incr_by=1):
        """Increment the key by the given amount."""
        return self.database.hincrby(self.key, key, incr_by)

    def incr_float(self, key, incr_by=1.):
        """Increment the key by the given amount."""
        return self.database.hincrbyfloat(self.key, key, incr_by)

    def as_dict(self, decode=False):
        """
        Return a dictionary containing all the key/value pairs in the
        hash.
        """
        res = self.database.hgetall(self.key)
        return decode_dict(res) if decode else res

    @classmethod
    def from_dict(cls, database, key, data, clear=False):
        """
        Create and populate a Hash object from a data dictionary.
        """
        hsh = cls(database, key)
        if clear:
            hsh.clear()
        hsh.update(data)
        return hsh


class List(Sortable, Container):
    """
    Redis List object wrapper. Supports a list-like interface.

    See `List commands <http://redis.io/commands#list>`_ for more info.
    """
    def __repr__(self):
        l = len(self)
        n_items = min(l, 10)
        items = safe_decode_list(self[:n_items])
        return '<List "%s": %s%s>' % (
            self.key,
            ', '.join(items),
            n_items < l and '...' or '')

    def __getitem__(self, item):
        """
        Retrieve an item from the list by index. In addition to
        integer indexes, you can also pass a ``slice``.
        """
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop
            if not stop:
                stop = -1
            else:
                stop -= 1
            return self.database.lrange(self.key, start, stop)
        return self.database.lindex(self.key, item)

    def __setitem__(self, idx, value):
        """Set the value of the given index."""
        return self.database.lset(self.key, idx, value)

    def __delitem__(self, item):
        """
        By default Redis treats deletes as delete by value, as
        opposed to delete by index. If an integer is passed into the
        function, it will be treated as an index, otherwise it will
        be treated as a value.

        If a slice is passed, then the list will be trimmed so that it *ONLY*
        contains the range specified by the slice start and stop. Note that
        this differs from the default behavior of Python's `list` type.
        """
        if isinstance(item, slice):
            start = item.start or 0
            stop = item.stop or -1
            if stop > 0:
                stop -= 1
            return self.database.ltrim(self.key, start, stop)
        elif isinstance(item, int):
            item = self[item]
            if item is None:
                return
        return self.database.lrem(self.key, 1, item)

    def __len__(self):
        """Return the length of the list."""
        return self.database.llen(self.key)

    def __iter__(self):
        """Iterate over the items in the list."""
        return iter(self.database.lrange(self.key, 0, -1))

    def append(self, value):
        """Add the given value to the end of the list."""
        return self.database.rpush(self.key, value)

    def prepend(self, value):
        """Add the given value to the beginning of the list."""
        return self.database.lpush(self.key, value)

    def extend(self, value):
        """Extend the list by the given value."""
        return self.database.rpush(self.key, *value)

    def insert(self, value, pivot, where):
        return self.database.linsert(self.key, where, pivot, value)

    def insert_before(self, value, key):
        """
        Insert the given value into the list before the index
        containing ``key``.
        """
        self.insert(value, key, 'before')

    def insert_after(self, value, key):
        """
        Insert the given value into the list after the index
        containing ``key``.
        """
        self.insert(value, key, 'after')

    def popleft(self):
        """Remove the first item from the list."""
        return self.database.lpop(self.key)

    def popright(self):
        """Remove the last item from the list."""
        return self.database.rpop(self.key)
    pop = popright

    def bpopleft(self, timeout=0):
        """
        Remove the first item from the list, blocking until an item becomes
        available or timeout is reached (0 for no timeout, default).
        """
        ret = self.database.blpop(self.key, timeout)
        if ret is not None:
            return ret[1]

    def bpopright(self, timeout=0):
        """
        Remove the last item from the list, blocking until an item becomes
        available or timeout is reached (0 for no timeout, default).
        """
        ret = self.database.brpop(self.key, timeout)
        if ret is not None:
            return ret[1]

    def move_tail(self, key):
        return self.database.rpoplpush(self.key, key)

    def bmove_tail(self, key, timeout=0):
        return self.database.brpoplpush(self.key, key, timeout)

    def as_list(self, decode=False):
        """
        Return a list containing all the items in the list.
        """
        items = self.database.lrange(self.key, 0, -1)
        return [_decode(item) for item in items] if decode else items

    @classmethod
    def from_list(cls, database, key, data, clear=False):
        """
        Create and populate a List object from a data list.
        """
        lst = cls(database, key)
        if clear:
            lst.clear()
        lst.extend(data)
        return lst


class Set(Sortable, Container):
    """
    Redis Set object wrapper. Supports a set-like interface.

    See `Set commands <http://redis.io/commands#set>`_ for more info.
    """
    def __repr__(self):
        return '<Set "%s": %s items>' % (self.key, len(self))

    def add(self, *items):
        """Add the given items to the set."""
        return self.database.sadd(self.key, *items)

    def __delitem__(self, item):
        """Remove the given item from the set."""
        return self.remove(item)

    def remove(self, *items):
        """Remove the given item(s) from the set."""
        return self.database.srem(self.key, *items)

    def pop(self):
        """Remove an element from the set."""
        return self.database.spop(self.key)

    def _first_or_any(self):
        return self.random()

    def __contains__(self, item):
        """
        Return a boolean value indicating whether the given item is
        a member of the set.
        """
        return self.database.sismember(self.key, item)

    def __len__(self):
        """Return the number of items in the set."""
        return self.database.scard(self.key)

    def _scan(self, *args, **kwargs):
        return self.database.sscan_iter(self.key, *args, **kwargs)

    def __iter__(self):
        """Return an iterable that yields the items of the set."""
        return iter(self._scan())

    def search(self, pattern, count=None):
        """
        Search the values of the given set using the specified pattern.

        :param str pattern: Pattern used to match keys.
        :param int count: Limit number of results returned.
        :returns: An iterator yielding matching values.
        """
        return self._scan(match=pattern, count=count)

    def members(self):
        """Return a ``set()`` containing the members of the set."""
        return self.database.smembers(self.key)

    def random(self, n=None):
        """Return a random member of the given set."""
        return self.database.srandmember(self.key, n)

    def __sub__(self, other):
        """
        Return the set difference of the current set and the left-
        hand :py:class:`Set` object.
        """
        return self.database.sdiff(self.key, other.key)

    def __or__(self, other):
        """
        Return the set union of the current set and the left-hand
        :py:class:`Set` object.
        """
        return self.database.sunion(self.key, other.key)

    def __and__(self, other):
        """
        Return the set intersection of the current set and the left-
        hand :py:class:`Set` object.
        """
        return self.database.sinter(self.key, other.key)

    @chainable_method
    def __isub__(self, other):
        self.diffstore(self.key, other)

    @chainable_method
    def __ior__(self, other):
        self.unionstore(self.key, other)

    @chainable_method
    def __iand__(self, other):
        self.interstore(self.key, other)

    def diffstore(self, dest, *others):
        """
        Store the set difference of the current set and one or more
        others in a new key.

        :param dest: the name of the key to store set difference
        :param others: One or more :py:class:`Set` instances
        :returns: A :py:class:`Set` referencing ``dest``.
        """
        keys = [self.key]
        keys.extend([other.key for other in others])
        self.database.sdiffstore(dest, keys)
        return self.database.Set(dest)

    def interstore(self, dest, *others):
        """
        Store the intersection of the current set and one or more
        others in a new key.

        :param dest: the name of the key to store intersection
        :param others: One or more :py:class:`Set` instances
        :returns: A :py:class:`Set` referencing ``dest``.
        """
        keys = [self.key]
        keys.extend([other.key for other in others])
        self.database.sinterstore(dest, keys)
        return self.database.Set(dest)

    def unionstore(self, dest, *others):
        """
        Store the union of the current set and one or more
        others in a new key.

        :param dest: the name of the key to store union
        :param others: One or more :py:class:`Set` instances
        :returns: A :py:class:`Set` referencing ``dest``.
        """
        keys = [self.key]
        keys.extend([other.key for other in others])
        self.database.sunionstore(dest, keys)
        return self.database.Set(dest)

    def as_set(self, decode=False):
        """
        Return a Python set containing all the items in the collection.
        """
        items = self.database.smembers(self.key)
        return set(_decode(item) for item in items) if decode else items

    @classmethod
    def from_set(cls, database, key, data, clear=False):
        """
        Create and populate a Set object from a data set.
        """
        s = cls(database, key)
        if clear:
            s.clear()
        s.add(*data)
        return s


class ZSet(Sortable, Container):
    """
    Redis ZSet object wrapper. Acts like a set and a dictionary.

    See `Sorted set commands <http://redis.io/commands#sorted_set>`_
    for more info.
    """
    def __repr__(self):
        l = len(self)
        n_items = min(l, 5)
        items = safe_decode_list(self[:n_items, False])
        return '<ZSet "%s": %s%s>' % (
            self.key,
            ', '.join(items),
            n_items < l and '...' or '')

    def add(self, _mapping=None, **kwargs):
        """
        Add the given item/score pairs to the ZSet. Arguments are
        specified as a dictionary of item: score, or as keyword arguments.
        """
        if _mapping is not None:
            _mapping.update(kwargs)
            mapping = _mapping
        else:
            mapping = _mapping
        return self.database.zadd(self.key, mapping)

    def _convert_slice(self, s):
        def _slice_to_indexes(s):
            start = s.start
            stop = s.stop
            if isinstance(start, int) or isinstance(stop, int):
                return start, stop
            if start:
                start = self.database.zrank(self.key, start)
                if start is None:
                    raise KeyError(s.start)
            if stop:
                stop = self.database.zrank(self.key, stop)
                if stop is None:
                    raise KeyError(s.stop)
            return start, stop
        start, stop = _slice_to_indexes(s)
        start = start or 0
        if not stop:
            stop = -1
        else:
            stop -= 1
        return start, stop

    def __getitem__(self, item):
        """
        Retrieve the given values from the sorted set. Accepts a
        variety of parameters for the input:

        .. code-block:: python

            zs = db.ZSet('my-zset')

            # Return the first 10 elements with their scores.
            zs[:10, True]

            # Return the first 10 elements without scores.
            zs[:10]
            zs[:10, False]

            # Return the range of values between 'k1' and 'k10' along
            # with their scores.
            zs['k1':'k10', True]

            # Return the range of items preceding and including 'k5'
            # without scores.
            zs[:'k5', False]
        """
        if isinstance(item, tuple) and len(item) == 2:
            item, withscores = item
        else:
            withscores = False

        if isinstance(item, slice):
            start, stop = self._convert_slice(item)
        else:
            start = stop = item

        return self.database.zrange(
            self.key,
            start,
            stop,
            withscores=withscores)

    def __setitem__(self, item, score):
        """Add item to the set with the given score."""
        return self.database.zadd(self.key, {item: score})

    def __delitem__(self, item):
        """
        Delete the given item(s) from the set. Like
        :py:meth:`~ZSet.__getitem__`, this method supports a wide
        variety of indexing and slicing options.
        """
        if isinstance(item, slice):
            start, stop = self._convert_slice(item)
            return self.database.zremrangebyrank(self.key, start, stop)
        else:
            return self.remove(item)

    def remove(self, *items):
        """Remove the given items from the ZSet."""
        return self.database.zrem(self.key, *items)

    def __contains__(self, item):
        """
        Return a boolean indicating whether the given item is in the
        sorted set.
        """
        return not (self.rank(item) is None)

    def __len__(self):
        """Return the number of items in the sorted set."""
        return self.database.zcard(self.key)

    def _scan(self, *args, **kwargs):
        return self.database.zscan_iter(self.key, *args, **kwargs)

    def __iter__(self):
        """
        Return an iterator that will yield (item, score) tuples.
        """
        return iter(self._scan())

    def iterator(self, with_scores=False, reverse=False):
        if with_scores and not reverse:
            return self.search(None)
        return self.range(
            0,
            -1,
            with_scores=with_scores,
            reverse=reverse)

    def search(self, pattern, count=None):
        """
        Search the set, returning items that match the given search
        pattern.

        :param str pattern: Search pattern using wildcards.
        :param int count: Limit result set size.
        :returns: Iterator that yields matching item/score tuples.
        """
        return self._scan(match=pattern, count=count)

    def score(self, item):
        """Return the score of the given item."""
        return self.database.zscore(self.key, item)

    def rank(self, item, reverse=False):
        """Return the rank of the given item."""
        fn = reverse and self.database.zrevrank or self.database.zrank
        return fn(self.key, item)

    def count(self, low, high=None):
        """
        Return the number of items between the given bounds.
        """
        if high is None:
            high = low
        return self.database.zcount(self.key, low, high)

    def lex_count(self, low, high):
        """
        Count the number of members in a sorted set between a given
        lexicographical range.
        """
        return self.database.zlexcount(self.key, low, high)

    def range(self, low, high, with_scores=False, desc=False, reverse=False):
        """
        Return a range of items between ``low`` and ``high``. By
        default scores will not be included, but this can be controlled
        via the ``with_scores`` parameter.

        :param low: Lower bound.
        :param high: Upper bound.
        :param bool with_scores: Whether the range should include the
            scores along with the items.
        :param bool desc: Whether to sort the results descendingly.
        :param bool reverse: Whether to select the range in reverse.
        """
        if reverse:
            return self.database.zrevrange(self.key, low, high, with_scores)
        else:
            return self.database.zrange(self.key, low, high, desc, with_scores)

    def range_by_score(self, low, high, start=None, num=None,
                       with_scores=False, reverse=False):
        if reverse:
            fn = self.database.zrevrangebyscore
            low, high = high, low
        else:
            fn = self.database.zrangebyscore
        return fn(self.key, low, high, start, num, with_scores)

    def range_by_lex(self, low, high, start=None, num=None, reverse=False):
        """
        Return a range of members in a sorted set, by lexicographical range.
        """
        if reverse:
            fn = self.database.zrevrangebylex
            low, high = high, low
        else:
            fn = self.database.zrangebylex
        return fn(self.key, low, high, start, num)

    def remove_by_rank(self, low, high=None):
        """
        Remove elements from the ZSet by their rank (relative position).

        :param low: Lower bound.
        :param high: Upper bound.
        """
        if high is None:
            high = low
        return self.database.zremrangebyrank(self.key, low, high)

    def remove_by_score(self, low, high=None):
        """
        Remove elements from the ZSet by their score.

        :param low: Lower bound.
        :param high: Upper bound.
        """
        if high is None:
            high = low
        return self.database.zremrangebyscore(self.key, low, high)

    def remove_by_lex(self, low, high):
        return self.database.zremrangebylex(self.key, low, high)

    def incr(self, key, incr_by=1.):
        """
        Increment the score of an item in the ZSet.

        :param key: Item to increment.
        :param incr_by: Amount to increment item's score.
        """
        return self.database.zincrby(self.key, incr_by, key)

    def _first_or_any(self):
        item = self[0]
        if item:
            return item[0]

    @chainable_method
    def __ior__(self, other):
        self.unionstore(self.key, other)
        return self

    @chainable_method
    def __iand__(self, other):
        self.interstore(self.key, other)
        return self

    def interstore(self, dest, *others, **kwargs):
        """
        Store the intersection of the current zset and one or more
        others in a new key.

        :param dest: the name of the key to store intersection
        :param others: One or more :py:class:`ZSet` instances
        :returns: A :py:class:`ZSet` referencing ``dest``.
        """
        keys = [self.key]
        keys.extend([other.key for other in others])
        self.database.zinterstore(dest, keys, **kwargs)
        return self.database.ZSet(dest)

    def unionstore(self, dest, *others, **kwargs):
        """
        Store the union of the current set and one or more
        others in a new key.

        :param dest: the name of the key to store union
        :param others: One or more :py:class:`ZSet` instances
        :returns: A :py:class:`ZSet` referencing ``dest``.
        """
        keys = [self.key]
        keys.extend([other.key for other in others])
        self.database.zunionstore(dest, keys, **kwargs)
        return self.database.ZSet(dest)

    def popmin(self, count=1):
        """
        Atomically remove the lowest-scoring item(s) in the set.

        :returns: a list of item, score tuples or ``None`` if the set is empty.
        """
        return self.database.zpopmin(self.key, count)

    def popmax(self, count=1):
        """
        Atomically remove the highest-scoring item(s) in the set.

        :returns: a list of item, score tuples or ``None`` if the set is empty.
        """
        return self.database.zpopmax(self.key, count)

    def bpopmin(self, timeout=0):
        """
        Atomically remove the lowest-scoring item from the set, blocking until
        an item becomes available or timeout is reached (0 for no timeout,
        default).

        Returns a 2-tuple of (item, score).
        """
        res = self.database.bzpopmin(self.key, timeout)
        if res is not None:
            return (res[1], res[2])

    def bpopmax(self, timeout=0):
        """
        Atomically remove the highest-scoring item from the set, blocking until
        an item becomes available or timeout is reached (0 for no timeout,
        default).

        Returns a 2-tuple of (item, score).
        """
        res = self.database.bzpopmax(self.key, timeout)
        if res is not None:
            return (res[1], res[2])

    def popmin_compat(self, count=1):
        """
        Atomically remove the lowest-scoring item(s) in the set. Compatible
        with Redis versions < 5.0.

        :returns: a list of item, score tuples or ``None`` if the set is empty.
        """
        pipe = self.database.pipeline()
        r1, r2 = (pipe
                  .zrange(self.key, 0, count - 1, withscores=True)
                  .zremrangebyrank(self.key, 0, count - 1)
                  .execute())
        return r1

    def popmax_compat(self, count=1):
        """
        Atomically remove the highest-scoring item(s) in the set. Compatible
        with Redis versions < 5.0.

        :returns: a list of item, score tuples or ``None`` if the set is empty.
        """
        pipe = self.database.pipeline()
        r1, r2 = (pipe
                  .zrange(self.key, 0 - count, -1, withscores=True)
                  .zremrangebyrank(self.key, 0 - count, -1)
                  .execute())
        return r1[::-1]

    def as_items(self, decode=False):
        """
        Return a list of 2-tuples consisting of key/score.
        """
        items = self.database.zrange(self.key, 0, -1, withscores=True)
        if decode:
            items = [(_decode(k), score) for k, score in items]
        return items

    @classmethod
    def from_dict(cls, database, key, data, clear=False):
        """
        Create and populate a ZSet object from a data dictionary.
        """
        zset = cls(database, key)
        if clear:
            zset.clear()
        zset.add(data)
        return zset


class HyperLogLog(Container):
    """
    Redis HyperLogLog object wrapper.

    See `HyperLogLog commands <http://redis.io/commands#hyperloglog>`_
    for more info.
    """
    def add(self, *items):
        """
        Add the given items to the HyperLogLog.
        """
        return self.database.pfadd(self.key, *items)

    def __len__(self):
        return self.database.pfcount(self.key)

    def __ior__(self, other):
        if not isinstance(other, (list, tuple)):
            other = [other]
        return self.merge(self.key, *other)

    def merge(self, dest, *others):
        """
        Merge one or more :py:class:`HyperLogLog` instances.

        :param dest: Key to store merged result.
        :param others: One or more ``HyperLogLog`` instances.
        """
        items = [self.key]
        items.extend([other.key for other in others])
        self.database.pfmerge(dest, *items)
        return HyperLogLog(self.database, dest)


class Array(Container):
    """
    Custom container that emulates an array (as opposed to the
    linked-list implementation of :py:class:`List`). This gives:

    * O(1) append, get, len, pop last, set
    * O(n) remove from middle

    :py:class:`Array` is built on top of the hash data type and
    is implemented using lua scripts.
    """
    def __getitem__(self, idx):
        """Get the value stored in the given index."""
        return self.database.run_script(
            'array_get',
            keys=[self.key],
            args=[idx])

    def __setitem__(self, idx, value):
        """Set the value at the given index."""
        return self.database.run_script(
            'array_set',
            keys=[self.key],
            args=[idx, value])

    def __delitem__(self, idx):
        """Delete the given index."""
        return self.pop(idx)

    def __len__(self):
        """Return the number of items in the array."""
        return self.database.hlen(self.key)

    def append(self, value):
        """Append a new value to the end of the array."""
        self.database.run_script(
            'array_append',
            keys=[self.key],
            args=[value])

    def extend(self, values):
        """Extend the array, appending the given values."""
        self.database.run_script(
            'array_extend',
            keys=[self.key],
            args=values)

    def pop(self, idx=None):
        """
        Remove an item from the array. By default this will be the
        last item by index, but any index can be specified.
        """
        if idx is not None:
            return self.database.run_script(
                'array_remove',
                keys=[self.key],
                args=[idx])
        else:
            return self.database.run_script(
                'array_pop',
                keys=[self.key],
                args=[])

    def __contains__(self, item):
        """
        Return a boolean indicating whether the given item is stored
        in the array. O(n).
        """
        item = encode(item)
        for value in self:
            if value == item:
                return True
        return False

    def __iter__(self):
        """Return an iterable that yields array items."""
        return iter(
            item[1] for item in sorted(self.database.hscan_iter(self.key)))

    def as_list(self, decode=False):
        """
        Return a list of items in the array.
        """
        return [_decode(i) for i in self] if decode else list(self)

    @classmethod
    def from_list(cls, database, key, data, clear=False):
        """
        Create and populate an Array object from a data dictionary.
        """
        arr = cls(database, key)
        if clear:
            arr.clear()
        arr.extend(data)
        return arr


def _normalize_stream_keys(keys, default_id='0-0'):
    if isinstance(keys, basestring_type):
        return {keys: default_id}
    elif isinstance(keys, (list, tuple)):
        return dict([(key, default_id) for key in keys])
    elif isinstance(keys, dict):
        return keys
    else:
        raise ValueError('keys must be either a stream key, a list of '
                         'stream keys, or a dictionary mapping key to '
                         'minimum message id.')


class Stream(Container):
    """
    Redis stream container.
    """
    def add(self, data, id='*', maxlen=None, approximate=True):
        """
        Add data to a stream.

        :param dict data: data to add to stream
        :param id: identifier for message ('*' to automatically append)
        :param maxlen: maximum length for stream
        :param approximate: allow stream max length to be approximate
        :returns: the added message id.
        """
        return self.database.xadd(self.key, data, id, maxlen, approximate)

    def __getitem__(self, item):
        """
        Read a range of values from a stream.

        The index must be a message id or a slice. An empty slice will result
        in reading all values from the stream. Message ids provided as lower or
        upper bounds are inclusive.

        To specify a maximum number of messages, use the "step" parameter of
        the slice.
        """
        if isinstance(item, slice):
            return self.range(item.start or '-', item.stop or '+', item.step)
        return self.get(item)

    def get(self, docid):
        """
        Get a message by id.

        :param docid: the message id to retrieve.
        :returns: a 2-tuple of (message id, data) or None if not found.
        """
        items = self[docid:docid:1]
        if items:
            return items[0]

    def range(self, start='-', stop='+', count=None):
        """
        Read a range of values from a stream.

        :param start: start key of range (inclusive) or '-' for oldest message
        :param stop: stop key of range (inclusive) or '+' for newest message
        :param count: limit number of messages returned
        """
        return self.database.xrange(self.key, start, stop, count)

    def revrange(self, start='+', stop='-', count=None):
        """
        Read a range of values from a stream in reverse.

        :param start: start key of range (inclusive) or '+' for newest message
        :param stop: stop key of range (inclusive) or '-' for oldest message
        :param count: limit number of messages returned
        """
        return self.database.xrevrange(self.key, start, stop, count)

    def __iter__(self):
        return iter(self[:])

    def __delitem__(self, item):
        """
        Delete one or more messages by id. The index can be either a single
        message id or a list/tuple of multiple ids.
        """
        if not isinstance(item, (list, tuple)):
            item = (item,)
        self.delete(*item)

    def delete(self, *id_list):
        """
        Delete one or more message by id. The index can be either a single
        message id or a list/tuple of multiple ids.
        """
        return self.database.xdel(self.key, *id_list)

    def length(self):
        """
        Return the length of a stream.
        """
        return self.database.xlen(self.key)
    __len__ = length

    def read(self, count=None, block=None, last_id=None):
        """
        Monitor stream for new data.

        :param int count: limit number of messages returned
        :param int block: milliseconds to block, 0 for indefinitely
        :param last_id: Last id read (an exclusive lower-bound). If the '$'
            value is given, we will only read values added *after* our command
            started blocking.
        :returns: a list of (message id, data) 2-tuples.
        """
        if last_id is None: last_id = '0-0'
        resp = self.database.xread({self.key: _decode(last_id)}, count, block)

            # resp is a 2-tuple of stream name -> message list.
        return resp[0][1] if resp else []

    def info(self):
        """
        Retrieve information about the stream. Wraps call to
        :py:meth:`~Database.xinfo_stream`.

        :returns: a dictionary containing stream metadata
        """
        return self.database.xinfo_stream(self.key)

    def groups_info(self):
        """
        Retrieve information about consumer groups for the stream. Wraps call
        to :py:meth:`~Database.xinfo_groups`.

        :returns: a dictionary containing consumer group metadata
        """
        return self.database.xinfo_groups(self.key)

    def consumers_info(self, group):
        """
        Retrieve information about consumers within the given consumer group
        operating on the stream. Calls :py:meth:`~Database.xinfo_consumers`.

        :param group: consumer group name
        :returns: a dictionary containing consumer metadata
        """
        return self.database.xinfo_consumers(self.key, group)

    def set_id(self, id):
        """
        Set the maximum message id for the stream.

        :param id: id of last-read message
        """
        return self.database.xsetid(self.key, id)

    def trim(self, count=None, approximate=True, minid=None, limit=None):
        """
        Trim the stream to the given "count" of messages, discarding the oldest
        messages first.

        :param count: maximum size of stream (maxlen)
        :param approximate: allow size to be approximate
        :param minid: evicts entries with IDs lower than the given min id.
        :param limit: maximum number of entries to evict.
        """
        return self.database.xtrim(self.key, maxlen=count,
                                   approximate=approximate, minid=minid,
                                   limit=limit)


class ConsumerGroupStream(Stream):
    """
    Helper for working with an individual stream within the context of a
    consumer group. This object is exposed as an attribute on a
    :py:class:`ConsumerGroup` object using the stream key for the attribute
    name.

    This class should not be created directly. It will automatically be added
    to the ``ConsumerGroup`` object.

    For example::

        cg = db.consumer_group('groupname', ['stream-1', 'stream-2'])
        cg.stream_1  # ConsumerGroupStream for "stream-1"
        cg.stream_2  # ConsumerGroupStream for "stream-2"
    """
    __slots__ = ('database', 'group', 'key', '_consumer')

    def __init__(self, database, group, key, consumer):
        self.database = database
        self.group = group
        self.key = key
        self._consumer = consumer

    def consumers_info(self):
        """
        Retrieve information about consumers within the given consumer group
        operating on the stream. Calls :py:meth:`~Database.xinfo_consumers`.

        :returns: a list of dictionaries containing consumer metadata
        """
        return self.database.xinfo_consumers(self.key, self.group)

    def ack(self, *id_list):
        """
        Acknowledge that the message(s) were been processed by the consumer
        associated with the parent :py:class:`ConsumerGroup`.

        :param id_list: one or more message ids to acknowledge
        :returns: number of messages marked acknowledged
        """
        return self.database.xack(self.key, self.group, *id_list)

    def claim(self, *id_list, **kwargs):
        """
        Claim pending - but unacknowledged - messages for this stream within
        the context of the parent :py:class:`ConsumerGroup`.

        :param id_list: one or more message ids to acknowledge
        :param min_idle_time: minimum idle time in milliseconds (keyword-arg).
        :returns: list of (message id, data) 2-tuples of messages that were
            successfully claimed
        """
        min_idle_time = kwargs.pop('min_idle_time', None) or 0
        if kwargs: raise ValueError('incorrect arguments for claim()')
        return self.database.xclaim(self.key, self.group, self._consumer,
                                    min_idle_time, id_list)

    def pending(self, start='-', stop='+', count=1000, consumer=None,
                idle=None):
        """
        List pending messages within the consumer group for this stream.

        :param start: start id (or '-' for oldest pending)
        :param stop: stop id (or '+' for newest pending)
        :param count: limit number of messages returned
        :param consumer: restrict message list to the given consumer
        :param int idle: filter by idle-time in milliseconds (6.2)
        :returns: A list containing status for each pending message. Each
            pending message returns [id, consumer, idle time, deliveries].
        """
        return self.database.xpending_range(self.key, self.group, min=start,
                                            max=stop, count=count,
                                            consumername=consumer, idle=idle)

    def autoclaim(self, consumer, min_idle_time, start_id=0, count=None, justid=False):
        """
        Transfer ownership of pending stream entries that match the specified
        criteria. Similar to calling XPENDING and XCLAIM, but provides a more
        straightforward way to deal with message delivery failures.

        :param consumer: name of consumer that claims the message.
        :param min_idle_time: in milliseconds
        :param start_id: start id
        :param count: optional, upper limit of entries to claim. Default 100.
        :param justid: return just IDs of messages claimed.
        :returns: [next start id, [messages that were claimed]
        """
        return self.database.xautoclaim(self.key, self.group, consumer,
                                        min_idle_time, start_id, count, justid)

    def read(self, count=None, block=None, last_id=None):
        """
        Monitor the stream for new messages within the context of the parent
        :py:class:`ConsumerGroup`.

        :param int count: limit number of messages returned
        :param int block: milliseconds to block, 0 for indefinitely.
        :param str last_id: optional last ID, by default uses the special
            token ">", which reads the oldest unread message.
        :returns: a list of (message id, data) 2-tuples.
        """
        key = {self.key: '>' if last_id is None else last_id}
        resp = self.database.xreadgroup(self.group, self._consumer, key, count,
                                        block)
        return resp[0][1] if resp else []

    def set_id(self, id='$'):
        """
        Set the last-read message id for the stream within the context of the
        parent :py:class:`ConsumerGroup`. By default this will be the special
        "$" identifier, meaning all messages are marked as having been read.

        :param id: id of last-read message (or "$").
        """
        return self.database.xgroup_setid(self.key, self.group, id)

    def delete_consumer(self, consumer=None):
        """
        Remove a specific consumer from a consumer group.

        :consumer: name of consumer to delete. If not provided, will be the
            default consumer for this stream.
        :returns: number of pending messages that the consumer had before
            being deleted.
        """
        if consumer is None: consumer = self._consumer
        return self.database.xgroup_delconsumer(self.key, self.group, consumer)


class ConsumerGroup(object):
    """
    Helper for working with Redis Streams consumer groups functionality. Each
    stream associated with the consumer group is exposed as a special attribute
    of the ``ConsumerGroup`` object, exposing stream-specific functionality
    within the context of the group.

    Rather than creating this class directly, use the
    :py:meth:`Database.consumer_group` method.

    Each registered stream within the group is exposed as a special attribute
    that provides stream-specific APIs within the context of the group. For
    more information see :py:class:`ConsumerGroupStream`.

    The streams managed by a consumer group must exist before the consumer
    group can be created. By default, calling :py:meth:`ConsumerGroup.create`
    will automatically create stream keys for any that do not exist.

    Example::

        cg = db.consumer_group('groupname', ['stream-1', 'stream-2'])
        cg.create()  # Create consumer group.
        cg.stream_1  # ConsumerGroupStream for "stream-1"
        cg.stream_2  # ConsumerGroupStream for "stream-2"
        # or, alternatively:
        cg.streams['stream-1']

    :param Database database: Redis client
    :param name: consumer group name
    :param keys: stream identifier(s) to monitor. May be a single stream
        key, a list of stream keys, or a key-to-minimum id mapping. The
        minimum id for each stream should be considered an exclusive
        lower-bound. The '$' value can also be used to only read values
        added *after* our command started blocking.
    :param consumer: name for consumer
    """
    stream_key_class = ConsumerGroupStream

    def __init__(self, database, name, keys, consumer=None):
        self.database = database
        self.name = name
        self.keys = _normalize_stream_keys(keys)
        self._read_keys = _normalize_stream_keys(list(self.keys), '>')
        self._consumer = consumer or (self.name + '.c1')
        self.streams = {}  # Dict of key->ConsumerGroupStream.

        # Add attributes for each stream exposed as part of the group.
        for key in self.keys:
            attr = make_python_attr(key)
            stream = self.stream_key_class(database, name, key, self._consumer)
            setattr(self, attr, stream)
            self.streams[key] = stream

    def consumer(self, name):
        """
        Create a new consumer for the :py:class:`ConsumerGroup`.

        :param name: name of consumer
        :returns: a :py:class:`ConsumerGroup` using the given consumer name.
        """
        return type(self)(self.database, self.name, self.keys, name)

    def create(self, ensure_keys_exist=True, mkstream=False):
        """
        Create the consumer group and register it with the group's stream keys.

        :param ensure_keys_exist: Ensure that the streams exist before creating
            the consumer group. Streams that do not exist will be created.
        :param mkstream: Use the "MKSTREAM" option to ensure stream exists (may
            require unstable version of Redis).
        """
        if ensure_keys_exist:
            for key in self.keys:
                if not self.database.exists(key):
                    msg_id = self.database.xadd(key, {'': ''}, id=b'0-1')
                    self.database.xdel(key, msg_id)
                elif self.database.type(key) != b'stream':
                    raise ValueError('Consumer group key "%s" exists and is '
                                     'not a stream. To prevent data-loss '
                                     'this key will not be deleted.' % key)

        resp = {}

        # Mapping of key -> last-read message ID.
        for key, value in self.keys.items():
            try:
                resp[key] = self.database.xgroup_create(key, self.name, value,
                                                        mkstream)
            except ResponseError as exc:
                if exception_message(exc).startswith('BUSYGROUP'):
                    resp[key] = False
                else:
                    raise
        return resp

    def reset(self):
        """
        Reset the consumer group, clearing the last-read status for each
        stream so it will read from the beginning of each stream.
        """
        return self.set_id('0-0')

    def destroy(self):
        """
        Destroy the consumer group.
        """
        resp = {}
        for key in self.keys:
            resp[key] = self.database.xgroup_destroy(key, self.name)
        return resp

    def read(self, count=None, block=None, consumer=None):
        """
        Read unseen messages from all streams in the consumer group. Wrapper
        for :py:class:`Database.xreadgroup` method.

        :param int count: limit number of messages returned
        :param int block: milliseconds to block, 0 for indefinitely.
        :param consumer: consumer name
        :returns: a list of (stream key, messages) tuples, where messages is
            a list of (message id, data) 2-tuples.
        """
        if consumer is None: consumer = self._consumer
        return self.database.xreadgroup(self.name, consumer, self._read_keys,
                                        count, block)

    def set_id(self, id='$'):
        """
        Set the last-read message id for each stream in the consumer group. By
        default, this will be the special "$" identifier, meaning all messages
        are marked as having been read.

        :param id: id of last-read message (or "$").
        """
        accum = {}
        for key in self.keys:
            accum[key] = self.database.xgroup_setid(key, self.name, id)
        return accum

    def stream_info(self):
        """
        Retrieve information for each stream managed by the consumer group.
        Calls :py:meth:`~Database.xinfo_stream` for each stream.

        :returns: a dictionary mapping stream key to a dictionary of metadata
        """
        accum = {}
        for key in self.keys:
            accum[key] = self.database.xinfo_stream(key)
        return accum


class BitFieldOperation(object):
    """
    Command builder for BITFIELD commands.
    """
    def __init__(self, database, key):
        self.database = database
        self.key = key
        self.operations = []
        self._last_overflow = None  # Default is "WRAP".

    def incrby(self, fmt, offset, increment, overflow=None):
        """
        Increment a bitfield by a given amount.

        :param fmt: format-string for the bitfield being updated, e.g. u8 for
            an unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :param int increment: value to increment the bitfield by.
        :param str overflow: overflow algorithm. Defaults to WRAP, but other
            acceptable values are SAT and FAIL. See the Redis docs for
            descriptions of these algorithms.
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        if overflow is not None and overflow != self._last_overflow:
            self._last_overflow = overflow
            self.operations.append(('OVERFLOW', overflow))

        self.operations.append(('INCRBY', fmt, offset, increment))
        return self

    def get(self, fmt, offset):
        """
        Get the value of a given bitfield.

        :param fmt: format-string for the bitfield being read, e.g. u8 for an
            unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        self.operations.append(('GET', fmt, offset))
        return self

    def set(self, fmt, offset, value):
        """
        Set the value of a given bitfield.

        :param fmt: format-string for the bitfield being read, e.g. u8 for an
            unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :param int value: value to set at the given position.
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        self.operations.append(('SET', fmt, offset, value))
        return self

    @property
    def command(self):
        return reduce(operator.add, self.operations, ('BITFIELD', self.key))

    def execute(self):
        """
        Execute the operation(s) in a single BITFIELD command. The return value
        is a list of values corresponding to each operation.
        """
        return self.database.execute_command(*self.command)

    def __iter__(self):
        """
        Implicit execution and iteration of the return values for a sequence of
        operations.
        """
        return iter(self.execute())


class BitField(Container):
    """
    Wrapper that provides a convenient API for constructing and executing Redis
    BITFIELD commands. The BITFIELD command can pack multiple operations into a
    single logical command, so the :py:class:`BitField` supports a method-
    chaining API that allows multiple operations to be performed atomically.

    Rather than instantiating this class directly, you should use the
    :py:meth:`Database.bit_field` method to obtain a :py:class:`BitField`.
    """
    def incrby(self, fmt, offset, increment, overflow=None):
        """
        Increment a bitfield by a given amount.

        :param fmt: format-string for the bitfield being updated, e.g. u8 for
            an unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :param int increment: value to increment the bitfield by.
        :param str overflow: overflow algorithm. Defaults to WRAP, but other
            acceptable values are SAT and FAIL. See the Redis docs for
            descriptions of these algorithms.
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        bfo = BitFieldOperation(self.database, self.key)
        return bfo.incrby(fmt, offset, increment, overflow)

    def get(self, fmt, offset):
        """
        Get the value of a given bitfield.

        :param fmt: format-string for the bitfield being read, e.g. u8 for an
            unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        bfo = BitFieldOperation(self.database, self.key)
        return bfo.get(fmt, offset)

    def set(self, fmt, offset, value):
        """
        Set the value of a given bitfield.

        :param fmt: format-string for the bitfield being read, e.g. u8 for an
            unsigned 8-bit integer.
        :param int offset: offset (in number of bits).
        :param int value: value to set at the given position.
        :returns: a :py:class:`BitFieldOperation` instance.
        """
        bfo = BitFieldOperation(self.database, self.key)
        return bfo.set(fmt, offset, value)

    def _validate_slice(self, item):
        if not isinstance(item, slice):
            raise ValueError('Must use a slice.')
        if item.stop is None or item.stop < 0:
            raise ValueError('slice must have a non-negative upper-bound')
        start = item.start or 0
        if start > item.stop:
            raise ValueError('start of slice cannot exceed stop')
        return start, item.stop

    def __getitem__(self, item):
        """
        Short-hand for getting a range of bits in a bitfield. Note that the
        item **must** be a slice specifying the start and end of the range of
        bits to read.
        """
        start, stop = self._validate_slice(item)
        return self.get('u%s' % (stop - start), start).execute()[0]

    def __setitem__(self, item, value):
        """
        Short-hand for setting a range of bits in a bitfield. Note that the
        item **must** be a slice specifying the start and end of the range of
        bits to read. If the value representation exceeds the number of bits
        implied by the slice range, a ``ValueError`` is raised.
        """
        start, stop = self._validate_slice(item)
        nbits = stop - start
        if value >= (1 << nbits):
            raise ValueError('value exceeds width specified by slice')
        self.set('u%s' % nbits, start, value).execute()

    def __delitem__(self, item):
        """
        Clear a range of bits in a bitfield. Note that the item **must** be a
        slice specifying the start and end of the range of bits to clear.
        """
        start, stop = self._validate_slice(item)
        self.set('u%s' % (stop - start), start, 0).execute()

    def get_raw(self):
        """
        Return the raw bytestring that comprises the bitfield. Equivalent to a
        normal GET command.
        """
        return self.database.get(self.key)

    def set_raw(self, value):
        """
        Set the raw bytestring that comprises the bitfield. Equivalent to a
        normal SET command.
        """
        return self.database.set(self.key, value)

    def bit_count(self, start=None, end=None):
        """
        Count the set bits in a string. Note that the `start` and `end`
        parameters are offsets in **bytes**.
        """
        return self.database.bitcount(self.key, start, end)

    def get_bit(self, offset):
        """
        Get the bit value at the given offset (in bits).

        :param int offset: bit offset
        :returns: value at bit offset, 1 or 0
        """
        return self.database.getbit(self.key, offset)

    def set_bit(self, offset, value):
        """
        Set the bit value at the given offset (in bits).

        :param int offset: bit offset
        :param int value: new value for bit, 1 or 0
        :returns: previous value at bit offset, 1 or 0
        """
        return self.database.setbit(self.key, offset, value)


class BloomFilter(Container):
    """
    Bloom-filters are probabilistic data-structures that are used to answer the
    question: "is X a member of set S?" It is possible to receive a false
    positive, but impossible to receive a false negative (in other words, if
    the bloom filter contains a value, it will never erroneously report that it
    does *not* contain such a value). The accuracy of the bloom-filter and the
    likelihood of a false positive can be reduced by increasing the size of the
    bloomfilter. The default size is 64KB (or 524,288 bits).

    Rather than instantiate this class directly, use
    :py:meth:`Database.bloom_filter`.
    """
    def __init__(self, database, key, size=64 * 1024):
        super(BloomFilter, self).__init__(database, key)
        self.size = size
        self.bits = self.size * 8
        self._bf = BitField(self.database, self.key)

    def _get_seeds(self, data):
        # Hash the data into a 16-byte digest, then break that up into 4 4-byte
        # (32-bit) unsigned integers. We use the modulo operator to normalize
        # these 32-bit ints to bit-indices.
        seeds = struct.unpack('>IIII', hashlib.md5(encode(data)).digest())
        return [seed % self.bits for seed in seeds]

    def add(self, data):
        """
        Add an item to the bloomfilter.

        :param bytes data: a bytestring representing the item to add.
        """
        bfo = BitFieldOperation(self.database, self.key)
        for bit_index in self._get_seeds(data):
            bfo.set('u1', bit_index, 1)
        bfo.execute()

    def contains(self, data):
        """
        Check if an item has been added to the bloomfilter.

        :param bytes data: a bytestring representing the item to check.
        :returns: a boolean indicating whether or not the item is present in
            the bloomfilter. False-positives are possible, but a negative
            return value is definitive.
        """
        bfo = BitFieldOperation(self.database, self.key)
        for bit_index in self._get_seeds(data):
            bfo.get('u1', bit_index)
        return all(bfo.execute())
    __contains__ = contains

    def __len__(self):
        return self.size
